package dev.pfaff.altargui.glfw;

import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMod;
import dev.pfaff.altargui.input.InputMods;
import dev.pfaff.glass.glfw.GLFWwindow;
import dev.pfaff.glass.glfw.KeyAction;
import dev.pfaff.glass.glfw.MouseButtonAction;
import dev.pfaff.glass.glfw.callbacks.*;
import dev.pfaff.log4truth.Logger;

import java.nio.file.Path;
import java.util.BitSet;
import java.util.List;

import static dev.pfaff.altargui.glfw.GLGlobals.GLFW;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.WARN;

public final class AltarGUIGLFWAdapter
	implements CharCallback, CursorPosCallback, FramebufferSizeCallback, KeyCallback, MouseButtonCallback,
			   PathDropCallback, ScrollCallback, WindowContentScaleCallback, WindowSizeCallback {
	private final GLFWwindow window;
	private final DelayedInitGuiContext gui;
	private int framebufferWidth, framebufferHeight;
	private int windowWidth, windowHeight;
	private float contentScaleX, contentScaleY;
	/**
	 * The scale factor necessary to convert screen coordinates (i.e. mouse position from GLFW callback) to "virtual
	 * pixels" used by Altar GUI.
	 */
	private float screenCoordinatesToVirtualPixelsScaleFactorX, screenCoordinatesToVirtualPixelsScaleFactorY;

	private double mouseX, mouseY;
	private final BitSet activeMouseButtons = new BitSet(8);

	public AltarGUIGLFWAdapter(GLFWwindow window, DelayedInitGuiContext gui) {
		this.window = window;
		this.gui = gui;
		assert window.equals(GLFWwindow.getCurrentContext(window.glfw()));
	}

	public void initializeState() {
		// This call is necessary for the following window.get* calls to return correct
		// information. For whatever reason, despite changing what those calls return, no events
		// will be fired as you might expect.
		GLFW.pollEvents();

		var fbSize = window.getFramebufferSize();
		var winSize = window.getWindowSize();
		var contentScale = window.getWindowContentScale();
		this.framebufferWidth = fbSize.width();
		this.framebufferHeight = fbSize.height();
		this.windowWidth = winSize.width();
		this.windowHeight = winSize.height();
		this.contentScaleX = contentScale.x();
		this.contentScaleY = contentScale.y();
		this.updateUI();
	}

	private void updateUI() {
		assert window.equals(GLFWwindow.getCurrentContext(window.glfw()));
		Logger.log(() -> STR."updateUI: framebufferWidth=\{framebufferWidth}, framebufferHeight=\{framebufferHeight}, contentScale=(\{contentScaleX}, \{contentScaleY})");
		this.recomputeVirtualScaleFactor();
		gui.onWindowResize(framebufferWidth, framebufferHeight, contentScaleX, contentScaleY);
	}

	private void recomputeVirtualScaleFactor() {
		this.screenCoordinatesToVirtualPixelsScaleFactorX = (float) (framebufferWidth / (double) windowWidth / contentScaleX);
		this.screenCoordinatesToVirtualPixelsScaleFactorY = (float) (framebufferHeight / (double) windowHeight / contentScaleY);
		Logger.log(() -> STR."screenCoordinatesToVirtualPixelsScaleFactor=(\{screenCoordinatesToVirtualPixelsScaleFactorX}, \{screenCoordinatesToVirtualPixelsScaleFactorY})");
	}

	private boolean checkOurWindowAndMainThread(GLFWwindow window) {
		if (!window.equals(this.window)) return false;
		assert window.equals(GLFWwindow.getCurrentContext(window.glfw()));
		return true;
	}

	@Override
	public void charTyped(GLFWwindow window, int codePoint) {
		if (!checkOurWindowAndMainThread(window)) return;

		gui.charTyped(codePoint);
	}

	@Override
	public void cursorMoved(GLFWwindow window, double xIn, double yIn) {
		if (!checkOurWindowAndMainThread(window)) return;

		var x = xIn * screenCoordinatesToVirtualPixelsScaleFactorX;
		var y = yIn * screenCoordinatesToVirtualPixelsScaleFactorY;

		var deltaX = x - mouseX;
		var deltaY = y - mouseY;
		mouseX = x;
		mouseY = y;

		gui.mouseMoved(x, y);
		activeMouseButtons.stream().forEach(button -> {
			gui.mouseDragged(x, y, button, deltaX, deltaY);
		});
	}

	@Override
	public void framebufferSizeChanged(GLFWwindow glfWwindow, int width, int height) {
		if (!checkOurWindowAndMainThread(window)) return;

		Logger.log(() -> STR."Received new framebuffer size: width=\{width}, height=\{height}", List.of(DEBUG));
		if (width == 0 || height == 0) {
			Logger.log(() -> "Ignoring width or height of 0", List.of(DEBUG));
			return;
		}

		this.framebufferWidth = width;
		this.framebufferHeight = height;
		this.updateUI();
	}

	@Override
	public void keyUpdated(GLFWwindow window, int keyCode, int scanCode, KeyAction action, int modifierBits) {
		if (!checkOurWindowAndMainThread(window)) return;

		// keyCode is -1 when a modifier key is pressed/released
		// TODO: add variants to InputKey for each modifier key and handle them by checking modifierBits and comparing
		//  with a stored "lastModifierBits"
		if (keyCode == -1) return;
		var key = InputKey.fromCode(keyCode);
		if (key == null) {
			Logger.log(() -> "Unrecognized key code " + keyCode, List.of(WARN));
			return;
		}
		var modifiers = InputMods.fromGLFW(modifierBits);
		if (action == KeyAction.RELEASE) {
			gui.keyReleased(key, scanCode, modifiers);
		} else {
			if (action == KeyAction.REPEAT) {
				modifiers = modifiers.set(InputMod.REPEAT);
			}
			gui.keyPressed(key, scanCode, modifiers);
		}
	}

	@Override
	public void mouseButtonUpdated(GLFWwindow window, int button, MouseButtonAction action, int modifiers) {
		if (!checkOurWindowAndMainThread(window)) return;

		if (action == MouseButtonAction.RELEASE) {
			activeMouseButtons.clear(button);
			gui.mouseReleased(mouseX, mouseY, button);
		} else {
			activeMouseButtons.set(button);
			gui.mousePressed(mouseX, mouseY, button);
		}
	}

	@Override
	public void pathDropped(GLFWwindow glfWwindow, List<Path> list) {
		if (!checkOurWindowAndMainThread(window)) return;

		// TODO: implement in AltarGUI
	}

	@Override
	public void scrolled(GLFWwindow window, double deltaX, double deltaY) {
		if (!checkOurWindowAndMainThread(window)) return;

		// TODO: use virtual scale factor
		gui.mouseScrolled(mouseX, mouseY, deltaX * screenCoordinatesToVirtualPixelsScaleFactorX, deltaY * screenCoordinatesToVirtualPixelsScaleFactorY);
	}

	@Override
	public void windowContentScaleChanged(GLFWwindow window, float x, float y) {
		if (!checkOurWindowAndMainThread(window)) return;

		Logger.log(() -> STR."Received new window content scale: x=\{x}, y=\{y}", List.of(DEBUG));
		this.contentScaleX = x;
		this.contentScaleY = y;
		this.updateUI();
	}

	@Override
	public void windowSizeChanged(GLFWwindow window, int width, int height) {
		if (!checkOurWindowAndMainThread(window)) return;

		Logger.log(() -> STR."Received new window size: width=\{width}, height=\{height}", List.of(DEBUG));
		if (width == 0 || height == 0) {
			Logger.log(() -> "Ignoring width or height of 0", List.of(DEBUG));
			return;
		}

		this.windowWidth = width;
		this.windowHeight = height;
		this.recomputeVirtualScaleFactor();
	}
}
